import 'dart:io';

import 'package:codenic_exception_converter/codenic_exception_converter.dart';
import 'package:mocktail/mocktail.dart';
import 'package:test/test.dart';

class MockLogger extends Mock implements CodenicLogger {}

class SocketFailure extends Failure {
  const SocketFailure({super.message = 'Socket failure occurred'});
}

class SocketExceptionConverter<T>
    extends ExceptionConverter<SocketException, SocketFailure, T> {
  const SocketExceptionConverter();

  @override
  SocketFailure convert({
    required SocketException exception,
    StackTrace? stackTrace,
    CodenicLogger? logger,
    MessageLog? messageLog,
  }) =>
      const SocketFailure();
}

void main() {
  group(
    'Exception converter',
    () {
      final messageLog = MessageLog(id: 'test');
      final exceptionConverterGroup = ExceptionConverterSuite();

      // setUp(() {
      //   reset(logger);
      // });

      group(
        'observe sync',
        () {
          test(
            'should return a `NetworkFailure` when a socket exception is '
            'thrown',
            () {
              exceptionConverterGroup
                  .observeSync<void>(
                    messageLog: messageLog,
                    exceptionConverters: [const SocketExceptionConverter()],
                    task: (messageLog) {
                      throw const SocketException('test');
                    },
                  )
                  .fold(
                    (l) => expect(l, isA<SocketFailure>()),
                    (r) => throw StateError(''),
                  );
            },
          );

          test(
            'should return a failure from the default exception converters '
            'when an exception is thrown',
            () {
              final exceptionConverter = ExceptionConverterSuite(
                exceptionConverters: [SocketExceptionConverter.new],
              );

              exceptionConverter
                  .observeSync<void>(
                    messageLog: messageLog,
                    task: (messageLog) {
                      throw const SocketException('test');
                    },
                  )
                  .fold(
                    (l) => expect(l, isA<SocketFailure>()),
                    (r) => throw StateError(''),
                  );
            },
          );

          test(
            'should return a base failure when an exception is thrown without '
            'an exception converter',
            () {
              exceptionConverterGroup
                  .observeSync<void>(
                    messageLog: messageLog,
                    task: (messageLog) {
                      throw const FormatException();
                    },
                  )
                  .fold(
                    (l) => expect(l, isA<Failure>()),
                    (r) => throw StateError(''),
                  );
            },
          );

          // test('should log a warning when a failure is returned, ', () {
          //   exceptionConverterGroup.observeSync<void>(
          //     messageLog: messageLog,
          //     task: (messageLog) {
          //       return const Left(Failure());
          //     },
          //   );

          //   verify(() => logger.warn(messageLog)).called(1);
          // });

          // test('should log an info on success', () {
          //   exceptionConverterGroup.observeSync<void>(
          //     messageLog: messageLog,
          //     task: (messageLog) {
          //       return const Right(null);
          //     },
          //   );

          //   verify(() => logger.info(messageLog)).called(1);
          // });
        },
      );

      group(
        'observe',
        () {
          test(
            'should return a `NetworkFailure` when a socket exception is '
            'thrown',
            () async {
              final result = await exceptionConverterGroup.observe<void>(
                messageLog: messageLog,
                exceptionConverters: [const SocketExceptionConverter()],
                task: (messageLog) {
                  throw const SocketException('test');
                },
              );

              result.fold(
                (l) => expect(l, isA<SocketFailure>()),
                (r) => throw StateError(''),
              );
            },
          );

          test(
            'should return a failure from the default exception converters '
            'when an exception is thrown',
            () async {
              final exceptionConverter = ExceptionConverterSuite(
                exceptionConverters: [SocketExceptionConverter.new],
              );

              final result = await exceptionConverter.observe<void>(
                messageLog: messageLog,
                task: (messageLog) {
                  throw const SocketException('test');
                },
              );

              result.fold(
                (l) => expect(l, isA<SocketFailure>()),
                (r) => throw StateError(''),
              );
            },
          );

          test(
            'should return a base failure when an exception is thrown without '
            'an exception converter',
            () async {
              final result = await exceptionConverterGroup.observe<void>(
                messageLog: messageLog,
                task: (messageLog) {
                  throw const FormatException();
                },
              );

              result.fold(
                (l) => expect(l, isA<Failure>()),
                (r) => throw StateError(''),
              );
            },
          );

          // test(
          //   'should log a warning when a failure is returned, ',
          //   () async {
          //     await exceptionConverterGroup.observe<void>(
          //       messageLog: messageLog,
          //       task: (messageLog) async {
          //         return const Left(Failure());
          //       },
          //     );

          //     verify(() => logger.warn(messageLog)).called(1);
          //   },
          // );

          // test(
          //   'should log an info on success',
          //   () async {
          //     await exceptionConverterGroup.observe<void>(
          //       messageLog: messageLog,
          //       task: (messageLog) async {
          //         return const Right(null);
          //       },
          //     );

          //     verify(() => logger.info(messageLog)).called(1);
          //   },
          // );
        },
      );

      group(
        'convert',
        () {
          test(
            'should convert `SocketException` to `NetworkFailure`',
            () {
              final result = exceptionConverterGroup.convert(
                messageLog: messageLog,
                error: const SocketException('test'),
                exceptionConverters: [const SocketExceptionConverter()],
              );

              expect(result, isA<SocketFailure>());
            },
          );

          test(
            'should convert `SocketException` to `NetworkFailure` using the '
            'default exception converters',
            () {
              final exceptionConverter = ExceptionConverterSuite(
                exceptionConverters: [SocketExceptionConverter.new],
              );

              final result = exceptionConverter.convert(
                messageLog: messageLog,
                error: const SocketException('test'),
              );

              expect(result, isA<SocketFailure>());
            },
          );

          test(
            'should throw an `Error` when an `Error` is passed',
            () {
              expect(
                () => exceptionConverterGroup.convert(
                  messageLog: messageLog,
                  error: const OutOfMemoryError(),
                ),
                throwsA(isA<OutOfMemoryError>()),
              );
            },
          );

          test(
            'should throw an `ArgumentError` when an a non-exception is passed',
            () {
              expect(
                () => exceptionConverterGroup.convert(
                  messageLog: messageLog,
                  error: 'Test',
                ),
                throwsA(isA<ArgumentError>()),
              );
            },
          );
        },
      );
    },
  );
}
